/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */

#include "ns3/socket.h"

#include "BitTorrentProtocol.h"

namespace ns3 {


BitTorrentProtocol::BitTorrentProtocol(void)
{
  m_protocolId = "19BitTorrent Protocol";
}

BitTorrentProtocol&
BitTorrentProtocol::GetInstance(void)
{
  static BitTorrentProtocol instance;
  return instance;
}

//====================================================================================================

bool
BitTorrentProtocol::SendHandshakeMessage(Ptr<Socket> socket, std::string peerId, std::string infoHash)
{
  std::string flags("00000000"); // Not to use DHT Algorithm

  std::stringstream ss;
  ss << m_protocolId;
  ss << flags;
  ss << infoHash;
  ss << peerId;

  std::string message = ss.str();

  int status = socket->Send((uint8_t*)message.data(), message.size(), 0);
  return (status != -1);
}



bool
BitTorrentProtocol::SendKeepAliveMessage(Ptr<Socket> socket)
{

  // The keep-alive message is constant. There is no payload.
  // Data Format : <len=0000>
  // Raw Data    : <0x00, 0x00, 0x00, 0x00>

  const uint32_t length = 0x00;

  std::stringstream ss;
  ss << getByteString(length);

  std::string message = ss.str();
  int status = socket->Send((uint8_t*)message.data(), message.size(), 0);
  return (status != -1);
}

bool
BitTorrentProtocol::SendChokeMessage(Ptr<Socket> socket)
{

  // The choke message is constant.
  // Data Format : <len=0001><id=0>
  // Raw Data    : <0x00, 0x00, 0x00, 0x01><0x00>

  const uint32_t length = 1;
  const uint8_t  id     = 0;

  std::stringstream ss;
  ss << getByteString(length);
  ss << getByteString(id);

  std::string message = ss.str();
  int status = socket->Send((uint8_t*)message.data(), message.size(), 0);
  return (status != -1);
}

bool
BitTorrentProtocol::SendUnchokeMessage(Ptr<Socket> socket)
{

  // The unchoke message is constant.
  // Data Format : <len=0001><id=1>
  // Raw Data    : <0x00, 0x00, 0x00, 0x01><0x01>

  const uint32_t length = 1;
  const uint8_t  id     = 1;

  std::stringstream ss;
  ss << getByteString(length);
  ss << getByteString(id);

  std::string message = ss.str();
  int status = socket->Send((uint8_t*)message.data(), message.size(), 0);
  return (status != -1);
}

bool
BitTorrentProtocol::SendInterestedMessage(Ptr<Socket> socket)
{

  // The interested message is constant.
  // Data Format : <len=0001><id=2>
  // Raw Data    : <0x00, 0x00, 0x00, 0x01><0x02>

  const uint32_t length = 1;
  const uint8_t  id     = 2;

  std::stringstream ss;
  ss << getByteString(length);
  ss << getByteString(id);

  std::string message = ss.str();
  int status = socket->Send((uint8_t*)message.data(), message.size(), 0);
  return (status != -1);
}

bool
BitTorrentProtocol::SendNotInterestedMessage(Ptr<Socket> socket)
{

  // The not interested message is constant.
  // Data Format : <len=0001><id=3>
  // Raw Data    : <0x00, 0x00, 0x00, 0x01><0x03>

  const uint32_t length = 1;
  const uint8_t  id     = 3;

  std::stringstream ss;
  ss << getByteString(length);
  ss << getByteString(id);

  std::string message = ss.str();
  int status = socket->Send((uint8_t*)message.data(), message.size(), 0);
  return (status != -1);
}

bool
BitTorrentProtocol::SendHaveMessage(Ptr<Socket> socket, const uint32_t pieceIndex)
{

  // The have message is fixed length.
  // Data Format : <len=0001><id=4>
  // Raw Data    : e.g.) <0x00, 0x00, 0x00, 0x01><0x04><0x00, 0x00, 0x00, 0x01>

  const uint32_t length = 1;
  const uint8_t  id     = 4;

  std::stringstream ss;
  ss << getByteString(length);
  ss << getByteString(id);
  ss << getByteString(pieceIndex);

  std::string message = ss.str();
  int status = socket->Send((uint8_t*)message.data(), message.size(), 0);
  return (status != -1);
}

bool
BitTorrentProtocol::SendBitfieldMessage(Ptr<Socket> socket, std::string bitfield)
{

  // The bitfield message is variable length.
  // Data Format : <len=0001+X><id=5><bitfield>
  // Raw Data    : e.g.) <0x00, 0x00, 0x00, 0x65><0x05><0xXX, 0xXX, 0xXX, ...>

  const uint32_t length = 1 + bitfield.size();
  const uint8_t  id     = 5;

  std::stringstream ss;
  ss << getByteString(length);
  ss << getByteString(id);
  ss << bitfield;

  std::string message = ss.str();
  int status = socket->Send((uint8_t*)message.data(), message.size(), 0);
  return (status != -1);
}

bool
BitTorrentProtocol::SendRequestMessage(Ptr<Socket> socket, uint32_t pieceIndex, uint32_t begin, uint32_t requestLength)
{

  // The request message is fixed length.
  // Data Format : <len=0013><id=6><index><begin><blockLength>
  // <index>         : the zero-based piece index
  // <begin>         : the zero-based byte offset within the piece
  // <requestLength> : the requested length
  // Raw Data : e.g.) <0x00, 0x00, 0x00, 0x0d><0x06>
  //                  <0x00, 0x00, 0x00, 0x01><0x00, 0x00, 0x00, 0x00><0x00, 0x00, 0x40, 0x00>

  const uint32_t length = 13;
  const uint8_t  id     = 6;

  std::stringstream ss;
  ss << getByteString(length);
  ss << getByteString(id);
  ss << getByteString(pieceIndex);
  ss << getByteString(begin);
  ss << getByteString(requestLength);

  std::string message = ss.str();
  int status = socket->Send((uint8_t*)message.data(), message.size(), 0);
  return (status != -1);
}


bool
BitTorrentProtocol::SendPieceMessage(Ptr<Socket> socket, uint32_t pieceIndex, uint32_t begin, std::string blockData)
{
  // The piece message is variable length.
  // Data Format : <len=009+X><id=7><index><begin><block>
  // <index> : the zero-based piece index
  // <begin> : the zero-based byte offset within the piece
  // <block> : block of date, which is a subset of the piece specified by index
  // Raw Data : e.g.) <0x00, 0x00, 0x40, 0x09><0x07>
  //                  <0x00, 0x00, 0x00, 0x01><0x00, 0x00, 0x00, 0x00><0xXX, 0xXX, 0xXX, ...>

  const uint32_t length = 9 + blockData.size();
  const uint8_t  id     = 7;

  std::stringstream ss;
  ss << getByteString(length);
  ss << getByteString(id);
  ss << getByteString(pieceIndex);
  ss << getByteString(begin);
  ss << blockData;

  std::string message = ss.str();
  int status = socket->Send((uint8_t*)message.data(), message.size(), 0);
  return (status != -1);
}

bool
BitTorrentProtocol::SendCancelMessage(Ptr<Socket> socket, uint32_t pieceIndex, uint32_t begin, uint32_t requestLength)
{
  // The cancel message is fixed length.
  // Data Format : <len=0013><id=8>
  // Raw Data    : <0x00, 0x00, 0x00, 0x0d><0x08>

  const uint32_t length = 13;
  const uint8_t  id     = 8;

  std::stringstream ss;
  ss << getByteString(length);
  ss << getByteString(id);
  ss << getByteString(pieceIndex);
  ss << getByteString(begin);
  ss << getByteString(requestLength);

  std::string message = ss.str();
  int status = socket->Send((uint8_t*)message.data(), message.size(), 0);
  return (status != -1);

}

//====================================================================================================

bool
BitTorrentProtocol::ParseHandshakeMessage(std::string message, std::string* peerId, std::string* infoHash)
{
  std::string id(message, 0, 21);
  std::string flags(message, 21, 8);

  *infoHash = std::string(message, 29, 20);
  *peerId   = std::string(message, 49, 20);

  if (id.compare(m_protocolId) == 0)
    {
      return true;
    }

  return false;
}

bool
BitTorrentProtocol::ParseHaveMessage(std::string message, uint32_t* pieceIndex)
{
  if (CheckMessageType(message) != BitTorrentProtocol::HAVE)
    return false;

  *pieceIndex = getUint32_t(std::string(message, 5, 4));
  return true;
}

bool
BitTorrentProtocol::ParseBitfieldMessage(std::string message, std::string* bitfield)
{
  if (CheckMessageType(message) != BitTorrentProtocol::BITFIELD)
    return false;

  *bitfield = std::string(message, 5, message.size() - 5);
  return true;
}

bool
BitTorrentProtocol::ParseRequestMessage(std::string message, uint32_t* pieceIndex, uint32_t* begin, uint32_t* requestLength)
{
  if (CheckMessageType(message) != BitTorrentProtocol::REQUEST)
    return false;

  *pieceIndex    = getUint32_t(std::string(message, 5, 4));
  *begin         = getUint32_t(std::string(message, 9, 4));
  *requestLength = getUint32_t(std::string(message, 13, 4));
  return true;
}

bool
BitTorrentProtocol::ParsePieceMessage(std::string message, uint32_t* pieceIndex, uint32_t* begin, std::string* blockData)
{
  if (CheckMessageType(message) != BitTorrentProtocol::PIECE)
    return false;

  *pieceIndex = getUint32_t(std::string(message, 5, 4));
  *begin      = getUint32_t(std::string(message, 9, 4));
  *blockData  = std::string(message, 13, message.size() - 13);
  return true;
}

bool
BitTorrentProtocol::ParseCancelMessage(std::string message, uint32_t* pieceIndex, uint32_t* begin, uint32_t* requestLength)
{
  if (CheckMessageType(message) != BitTorrentProtocol::CANCEL)
    return false;

  *pieceIndex    = getUint32_t(std::string(message, 5, 4));
  *begin         = getUint32_t(std::string(message, 9, 4));
  *requestLength = getUint32_t(std::string(message, 13, message.size() - 13));
  return true;
}

//====================================================================================================

BitTorrentProtocol::MessageType
BitTorrentProtocol::CheckMessageType(std::string message)
{
  uint32_t idealSize = getIdealMessageSize(message);
  if (message.size() < idealSize)
    return INCOMPLETE;

  if (idealSize == 4)
    return KEEP_ALIVE;

  BitTorrentProtocol::MessageType id = (MessageType)(*std::string(message, 4, 1).data());
  return id;
}

uint32_t
BitTorrentProtocol::getIdealMessageSize(std::string message)
{
  const uint32_t HEADER = 4;
  uint32_t length = getUint32_t(std::string(message, 0, 4));

  return (HEADER + length);
}

//====================================================================================================

std::string
BitTorrentProtocol::getByteString(uint32_t value)
{
  std::stringstream ss;
  ss << (char)((value >> 24) & 0xff);
  ss << (char)((value >> 16) & 0xff);
  ss << (char)((value >>  8) & 0xff);
  ss << (char)((value      ) & 0xff);
  return ss.str();
}

std::string
BitTorrentProtocol::getByteString(uint8_t value)
{
  std::stringstream ss;
  ss << (char)value;
  return ss.str();
}

uint32_t
BitTorrentProtocol::getUint32_t(std::string data)
{
  uint32_t value = 0x00;
  std::string::iterator it = data.begin();
  for (int i=0; i<4; i++)
    {
      value |= (uint8_t)*it;
      if (++it != data.end())
        value <<= 8;
      else
        break;
    }
  return value;
}

}

