/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */

#include <vector>
#include <math.h>

#include "ns3/log.h"

#include "ns3/MetainfoFile.h"
#include "ns3/PeerInformation.h"
#include "ns3/ConnectionManager.h"

#include "ClientDatabase.h"

namespace ns3 {

NS_LOG_COMPONENT_DEFINE("ClientDatabase");
NS_OBJECT_ENSURE_REGISTERED(ClientDatabase);

ClientDatabase::ClientDatabase(Ptr<MetainfoFile> metainfo, bool isSeeder)
{
  NS_LOG_FUNCTION(this);

  m_infoHash    = metainfo->infoHash();
  m_fileLength  = metainfo->fileLength();
  m_pieceLength = metainfo->pieceLength();

  m_pieceIndexLength = ceil((double)m_fileLength/(double)m_pieceLength);

  if (isSeeder)
    {
      m_status = ClientDatabase::COMPLETED;
      m_left = 0;
      m_bitfield.assign(m_pieceIndexLength, true);
    }
  else
    {
      m_status = ClientDatabase::STARTED;
      m_left = m_fileLength;
      m_bitfield.assign(m_pieceIndexLength, false);
    }

  m_listeningPort = 6881;

  // Set default AS number
  setAsn(0);
}

void
ClientDatabase::setConnectionManager(Ptr<ConnectionManager> connectionManager)
{
  //TODO: replace observer design pattern
  m_connectionManager = connectionManager;
}

std::string
ClientDatabase::getPeerId(void) const
{
  return m_peerId;
}

void
ClientDatabase::setPeerId(std::string peerId)
{
  m_peerId = peerId;
}

std::string
ClientDatabase::getInfoHash(void) const
{
  return m_infoHash;
}

ClientDatabase::Status
ClientDatabase::getStatus(void) const
{
  return m_status;
}

uint32_t
ClientDatabase::getPieceLength(void) const
{
  return m_pieceLength;
}

void
ClientDatabase::setListeningPort(uint16_t port)
{
  m_listeningPort = port;
}

uint16_t
ClientDatabase::listeningPort(void)
{
  return m_listeningPort;
}

uint32_t
ClientDatabase::uploaded(void)
{
  return m_uploaded;
}

uint32_t
ClientDatabase::downloaded(void)
{
  return m_downloaded;
}

uint32_t
ClientDatabase::left(void)
{
  return m_left;
}

const std::vector<bool>&
ClientDatabase::getBitfield(void) const
{
  return m_bitfield;
}

std::string
ClientDatabase::getBitfieldString(void) const
{
  std::stringstream ss;

  for (auto it = m_bitfield.begin(); it != m_bitfield.end(); )
    {
      uint8_t data = 0x00;
      for (int i=0; i<8 && (it != m_bitfield.end()); i++)
        {
          data |= ((*it++ ? 0x01 : 0x00) << (7 - i));
        }
      ss << (char)data;
    }

  return ss.str();
}

bool
ClientDatabase::havePiece(uint32_t pieceIndex) const
{
  return m_bitfield.at(pieceIndex);
}

bool
ClientDatabase::receiveNewPiece(uint32_t pieceIndex, uint32_t begin, const std::string& blockData)
{
  NS_LOG_FUNCTION(this);

  if (havePiece(pieceIndex))
    return false;

  if ((begin != 0) || (blockData.size() != m_pieceLength))
    return false;

  m_bitfield.at(pieceIndex) = true;
  UpdateStatus();

  //DISCUSSION: should change to Observer design pattern.
  if (m_connectionManager != NULL)
    {
      m_connectionManager->NotifiedGetNewPiece(pieceIndex);
    }
  return true;
}

void
ClientDatabase::UpdateStatus(void)
{
  if (m_status == ClientDatabase::STOPPED)
    return;

  for (std::vector<bool>::iterator it = m_bitfield.begin();;)
    {
      if (*it++ == false)
        {
          m_status = ClientDatabase::STARTED;
          break;
        }

      if(it == m_bitfield.end())
        {
          m_status = ClientDatabase::COMPLETED;
          break;
        }
    }
}


//====================================================================================================

void
ClientDatabase::setAsn(uint16_t asn)
{
  m_asn = asn;
}

uint16_t
ClientDatabase::getAsn(void) const
{
  return m_asn;
}


//====================================================================================================

void
ClientDatabase::ClearPeerList(void)
{
  m_peerList.clear();
}

void
ClientDatabase::AddToPeerList(std::string peer_id, std::string ip, uint16_t port, uint16_t asn)
{
  if (getPeerId().compare(peer_id) != 0)
    {
      PeerInformation pi(peer_id, ip, port);
      pi.setAsn(asn);
      
      m_peerList.push_back(pi);
    }
}

const std::vector<PeerInformation>&
ClientDatabase::getPeerList(void) const
{
  return m_peerList;
}



}

