/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */

#include "ns3/log.h"
#include "ns3/simulator.h"

#include "ns3/BitTorrentClient.h"
#include "ns3/ClientDatabase.h"
#include "ns3/PeerInformation.h"

#include "ns3/ListeningPeerConnection.h"

#include "ns3/FullRandomConnectionStrategy.h"
#include "ns3/FullRandomUnchokeStrategy.h"
#include "ns3/SimpleChokeStrategy.h"

#include "ConnectionManager.h"

namespace ns3 {

NS_LOG_COMPONENT_DEFINE("BitTorrentConnectionManager");
NS_OBJECT_ENSURE_REGISTERED(ConnectionManager);

ConnectionManager::ConnectionManager(Ptr<BitTorrentClient> client, Ptr<ClientDatabase> clientDB)
{
  NS_LOG_FUNCTION(this);

  m_client   = client;
  m_clientDB = clientDB;

  m_clientDB->setConnectionManager(this);

  m_listeningPeerConnection = Create<ListeningPeerConnection>(this);

  // set default value
  m_nConnectionMax = 32;
  m_nUnchokeMax    = 4;

  // set default strategy
  m_connectionStrategy = Create<FullRandomConnectionStrategy>(this);
  m_chokeStrategy      = Create<SimpleChokeStrategy>(this);
  m_unchokeStrategy    = Create<FullRandomUnchokeStrategy>(this);
}

bool
ConnectionManager::StartConnectionManager(void)
{
  NS_LOG_FUNCTION(this);

  uint16_t port = m_clientDB->listeningPort();
  m_listeningPeerConnection->StartListening(port);

  m_connectionStrategy->StartConnect();
  m_chokeStrategy->StartChoke();
  m_unchokeStrategy->StartUnchoke();

  return true;
}

bool
ConnectionManager::StopConnectionManager(void)
{
  NS_LOG_FUNCTION(this);

  m_listeningPeerConnection->StopListening();

  m_connectionStrategy->StopConnect();
  m_chokeStrategy->StopChoke();
  m_unchokeStrategy->StopUnchoke();

  CloseAllConnection();

  return true;
}

//====================================================================================================

Ptr<Socket>
ConnectionManager::getListeningSocket(void)
{
  NS_LOG_FUNCTION(this);

  if (!m_listeningSocket)
    {
      m_listeningSocket = m_client->CreateTcpSocket();
    }

  return m_listeningSocket;
}

Ipv4Address
ConnectionManager::getListeningIpv4Address(void)
{
  NS_LOG_FUNCTION(this);

  return m_client->ClientIpv4Address();
}

Ptr<Socket>
ConnectionManager::CreateTcpSocket(void)
{
  NS_LOG_FUNCTION(this);

  return m_client->CreateTcpSocket();
}

bool
ConnectionManager::isConnected(std::string peerId)
{
  NS_LOG_FUNCTION(this);

  for (std::list<Ptr<PeerConnection> >::iterator it = m_connectedPeerList.begin();
       it != m_connectedPeerList.end(); ++it)
    {
      std::string connectedPeerId = (*it)->getPeerId();
      if (connectedPeerId.compare(peerId) == 0)
        return true;
    }
  return false;
}

uint32_t
ConnectionManager::RestConnectableSocket(void)
{
  NS_LOG_FUNCTION(this);

  return m_nConnectionMax - m_connectedPeerList.size();
}

//====================================================================================================

bool
ConnectionManager::isAcceptableNewConnection(void)
{
  NS_LOG_FUNCTION(this);

  return (m_connectedPeerList.size() < m_nConnectionMax);
}

bool
ConnectionManager::EntrustedPeerConnection(Ptr<Socket> socket, std::string peerId)
{
  NS_LOG_FUNCTION(this);

  if (isAcceptableNewConnection())
    {
      Ptr<PeerConnection> pc = Create<PeerConnection>(socket, peerId, this);
      m_connectedPeerList.push_back(pc);

      std::stringstream ss;
      ss << "Entrusted a Handshaked Connection [" << peerId << "]";
      ss << " / ";
      ss << "Connected socket [" << m_connectedPeerList.size() << "/" << m_nConnectionMax << "]";
      CONNECTION_MANAGER_LOG_INFO(ss.str());
    }
  return false;;
}

void
ConnectionManager::RemoveFromPeerList(Ptr<PeerConnection> peerConnection)
{
  NS_LOG_FUNCTION(this);

  m_connectedPeerList.remove(peerConnection);
  m_unchokingPeerList.remove(peerConnection);
}

//====================================================================================================

void
ConnectionManager::ReceiveNewPiece(uint32_t pieceIndex, uint32_t begin, const std::string& message)
{
  NS_LOG_FUNCTION(this);

  m_clientDB->receiveNewPiece(pieceIndex, begin, message);
}

void
ConnectionManager::NotifiedGetNewPiece(uint32_t pieceIndex)
{
  NS_LOG_FUNCTION(this);

  BroadcastHaveMessage(pieceIndex);
}

void
ConnectionManager::BroadcastHaveMessage(uint32_t pieceIndex)
{
  NS_LOG_FUNCTION(this);

  for (std::list<Ptr<PeerConnection> >::iterator it = m_connectedPeerList.begin();
       it != m_connectedPeerList.end(); ++it)
    {
      (*it)->SendHaveMessage(pieceIndex);
    }
}

void
ConnectionManager::CloseAllConnection(void)
{
  NS_LOG_FUNCTION(this);

  for (auto it = m_unchokingPeerList.begin();
       it != m_unchokingPeerList.end(); )
    {
      Choke(*it++);
    }

  for (auto it = m_connectedPeerList.begin();
       it != m_connectedPeerList.end(); )
    {
      (*it)->CloseConnection();
      m_connectedPeerList.remove(*it++);
    }

}

//====================================================================================================

bool
ConnectionManager::Choke(Ptr<PeerConnection> peer)
{
  NS_LOG_FUNCTION(this);

  if (!isUnchoking(peer))
    return false;
  
  peer->SendChokeMessage();
  m_unchokingPeerList.remove(peer);

  std::stringstream ss;
  ss << "Choke [" << peer->getPeerInformation()->getPeerId() << "]";
  ss << " / ";
  ss << "Unchoking status [" << m_unchokingPeerList.size() << "/" << m_nUnchokeMax << "]";
  CONNECTION_MANAGER_LOG_INFO(ss.str());

  return true;
}

uint32_t
ConnectionManager::UnchokeMax(void) const
{
  NS_LOG_FUNCTION(this);

  return m_nUnchokeMax;
}

bool
ConnectionManager::isUnchoking(Ptr<PeerConnection> peer)
{
  NS_LOG_FUNCTION(this);

  for (std::list<Ptr<PeerConnection> >::iterator it = m_unchokingPeerList.begin();
       it != m_unchokingPeerList.end(); ++it)
    {
      if(*it == peer)
        return true;
    }
  return false;
}

bool
ConnectionManager::isUnchokable(Ptr<PeerConnection> peer)
{
  NS_LOG_FUNCTION(this);

  if (isUnchoking(peer))
    {
      return false;
    }

  if (!(m_unchokingPeerList.size() < m_nUnchokeMax))
    {
      return false;
    }

  return true;
}

bool
ConnectionManager::Unchoke(Ptr<PeerConnection> peer)
{
  NS_LOG_FUNCTION(this);

  if (isUnchokable(peer))
    {
      peer->SendUnchokeMessage();
      m_unchokingPeerList.push_back(peer);

      std::stringstream ss;
      ss << "Unchoke [" << peer->getPeerInformation()->getPeerId() << "]";
      ss << " / ";
      ss << "Unchoking status [" << m_unchokingPeerList.size() << "/" << m_nUnchokeMax << "]";
      CONNECTION_MANAGER_LOG_INFO(ss.str());

      return true;
    }
  return false;
}

//====================================================================================================

const Ptr<ClientDatabase>&
ConnectionManager::getClientDatabase(void) const
{
  NS_LOG_FUNCTION(this);

  return m_clientDB;
}

const std::list<Ptr<PeerConnection> >&
ConnectionManager::getConnectedPeerList(void) const
{
  NS_LOG_FUNCTION(this);

  return m_connectedPeerList;
}

const std::list<Ptr<PeerConnection> >&
ConnectionManager::getUnchokingPeerList(void) const
{
  NS_LOG_FUNCTION(this);

  return m_unchokingPeerList;
}

//====================================================================================================

void
ConnectionManager::CONNECTION_MANAGER_LOG_INFO(std::string info)
{
  NS_LOG_INFO("[" << m_clientDB->getPeerId() << "] " <<
              Simulator::Now().GetSeconds() << "[s] " << info);
}

}

