/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */

#include "ns3/log.h"
#include "ns3/simulator.h"
#include "ns3/socket.h"
#include "ns3/packet.h"

#include "ns3/BitTorrentProtocol.h"
#include "ns3/ConnectionManager.h"
#include "ns3/ClientDatabase.h"

#include "ListeningPeerConnection.h"

namespace ns3 {

NS_LOG_COMPONENT_DEFINE("ListeningPeerConnection");
NS_OBJECT_ENSURE_REGISTERED(ListeningPeerConnection);

ListeningPeerConnection::ListeningPeerConnection(Ptr<ConnectionManager> connectionManager)
{
  NS_LOG_FUNCTION(this);

  m_connectionManager = connectionManager;
  m_nConnection = 0;
}

bool
ListeningPeerConnection::StartListening(uint16_t port)
{
  NS_LOG_FUNCTION(this);

  if (!m_socket)
    {
      m_socket = m_connectionManager->getListeningSocket();
      Ipv4Address       ipv4Address   = m_connectionManager->getListeningIpv4Address();
      InetSocketAddress socketAddress = InetSocketAddress(ipv4Address, port);

      m_socket->Bind(socketAddress);
      m_socket->Listen();
    }

  m_socket->SetRecvCallback(MakeCallback(&ListeningPeerConnection::HandleRead, this));
  m_socket->SetAcceptCallback(
    MakeNullCallback<bool, Ptr<Socket>, const Address &>(),
    MakeCallback(&ListeningPeerConnection::HandleAccept, this));
  m_socket->SetCloseCallbacks(
    MakeCallback(&ListeningPeerConnection::HandlePeerClose, this),
    MakeCallback(&ListeningPeerConnection::HandlePeerError, this));
}

bool
ListeningPeerConnection::StopListening(void)
{
  NS_LOG_FUNCTION(this);

  for (std::map<Ptr<Socket>, Ptr<Packet> >::iterator it = m_acceptedSocketList.begin();
       it != m_acceptedSocketList.end(); ) // Don't increment it here.
    {
      it->first->Close();
      it->first->SetRecvCallback(MakeNullCallback<void, Ptr<Socket> >());

      m_nConnection--;
      m_acceptedSocketList.erase(it++);
    }

  if (m_socket)
    {
      m_socket->Close();
      m_socket->SetRecvCallback(MakeNullCallback<void, Ptr<Socket> >());
      m_socket = 0;
    }
}

void
ListeningPeerConnection::HandleAccept(Ptr<Socket> socket, const Address& from)
{
  NS_LOG_FUNCTION(this << socket << from);

  m_nConnection++;
  socket->SetRecvCallback(MakeCallback(&ListeningPeerConnection::HandleRead, this));
  m_acceptedSocketList[socket] = Create<Packet>();
}

void
ListeningPeerConnection::HandleRead(Ptr<Socket> socket)
{
  NS_LOG_FUNCTION(this);

  Ptr<Packet> packet = m_acceptedSocketList[socket];

  for (uint32_t available = socket->GetRxAvailable();
       available > 0; available = socket->GetRxAvailable())
    {
      packet->AddAtEnd(socket->Recv());
    }

  uint8_t *buf = new uint8_t[packet->GetSize()];
  packet->CopyData(buf, packet->GetSize());
  std::string recvData((char*)buf, packet->GetSize());
  delete[] buf;

  //TODO: check the packet size
  // insert code here....

  BitTorrentProtocol bPro = BitTorrentProtocol::GetInstance();
  std::string peerId;
  std::string infoHash;

  if (bPro.ParseHandshakeMessage(recvData, &peerId, &infoHash) &&
      (infoHash.compare(m_connectionManager->getClientDatabase()->getInfoHash()) == 0))
    {
      bPro.SendHandshakeMessage(socket, m_connectionManager->getClientDatabase()->getPeerId(), infoHash);
      EntrustPeerConnection(socket, peerId);
    }
  else
    {
      socket->Close();

      m_nConnection--;
      m_acceptedSocketList.erase(socket);
    }
}

void
ListeningPeerConnection::HandlePeerClose(Ptr<Socket> socket)
{
  NS_LOG_FUNCTION(this);
}

void
ListeningPeerConnection::HandlePeerError(Ptr<Socket> socket)
{
  NS_LOG_FUNCTION(this);
}

void
ListeningPeerConnection::EntrustPeerConnection(Ptr<Socket> socket, std::string peerId)
{
  NS_LOG_FUNCTION(this);

  socket->SetRecvCallback(
    MakeNullCallback<void, Ptr<Socket> >());

  m_nConnection--;
  m_acceptedSocketList.erase(socket);

  m_connectionManager->EntrustedPeerConnection(socket, peerId);

  std::stringstream ss;
  ss << "++++++++++ Handshake with [" << peerId << "] Succeeded!! (Listening) ++++++++++";
  LISTENING_LOG_INFO(ss.str());
}

void
ListeningPeerConnection::LISTENING_LOG_INFO(std::string info)
{
  NS_LOG_INFO("[" << m_connectionManager->getClientDatabase()->getPeerId() << "] " <<
              Simulator::Now().GetSeconds() << "[s] " << info);
}


}

