/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */

#include <random>

#include "ns3/log.h"
#include "ns3/socket.h"
#include "ns3/packet.h"

#include "ns3/BitTorrentProtocol.h"
#include "ns3/FullRandomRequestStrategy.h"

#include "PeerConnection.h"

namespace ns3 {

NS_LOG_COMPONENT_DEFINE("PeerConnection");
NS_OBJECT_ENSURE_REGISTERED(PeerConnection);

PeerConnection::PeerConnection(Ptr<Socket> socket, std::string peerId, Ptr<ConnectionManager> connectionManager)
{
  NS_LOG_FUNCTION(this);

  m_connectionManager = connectionManager;
  m_socket            = socket;
  m_packet            = Create<Packet>();
  m_peerInfo          = Create<PeerInformation>(peerId);

  m_socket->SetRecvCallback(
    MakeCallback(&PeerConnection::HandleRead, this));
  m_socket->SetCloseCallbacks(
    MakeCallback(&PeerConnection::HandlePeerClose, this),
    MakeCallback(&PeerConnection::HandlePeerError, this));

  BitTorrentProtocol::GetInstance().SendBitfieldMessage(socket, m_connectionManager->getClientDatabase()->getBitfieldString());

  // set default strategy
  m_requestStrategy = Create<FullRandomRequestStrategy>(this);
}

void
PeerConnection::HandleRead(Ptr<Socket> socket)
{
  //  NS_LOG_FUNCTION(this);

  for (uint32_t available = socket->GetRxAvailable();
       available > 0; available = socket->GetRxAvailable())
    {
      m_packet->AddAtEnd(socket->Recv());
    }

  uint8_t *buf = new uint8_t[m_packet->GetSize()];
  m_packet->CopyData(buf, m_packet->GetSize());
  std::string recvData((char*)buf, m_packet->GetSize());
  delete[] buf;

  BitTorrentProtocol bPro = BitTorrentProtocol::GetInstance();

  if (bPro.CheckMessageType(recvData) == BitTorrentProtocol::INCOMPLETE)
    return; // continue to receive message until complete state come

  std::stringstream ss;
  ss << "Received a message: " << recvData.size() << " byte";
  PEER_CONNECTION_LOG_INFO(ss.str());

  switch (bPro.CheckMessageType(recvData))
    {
    case BitTorrentProtocol::CHOKE:
      RecvChokeMessage(recvData);
      break;
    case BitTorrentProtocol::UNCHOKE:
      RecvUnchokeMessage(recvData);
      break;
    case BitTorrentProtocol::INTERESTED:
      RecvInterestedMessage(recvData);
      break;
    case BitTorrentProtocol::NOT_INTERESTED:
      RecvNotInterestedMessage(recvData);
      break;
    case BitTorrentProtocol::HAVE:
      RecvHaveMessage(recvData);
      break;
    case BitTorrentProtocol::BITFIELD:
      RecvBitfieldMessage(recvData);
      break;
    case BitTorrentProtocol::REQUEST:
      RecvRequestMessage(socket, recvData);
      break;
    case BitTorrentProtocol::PIECE:
      RecvPieceMessage(recvData);
      break;
    case BitTorrentProtocol::CANCEL:
      RecvCancelMessage(recvData);
      break;
    case BitTorrentProtocol::KEEP_ALIVE:
      RecvKeepAliveMessage(socket, recvData);
      break;
    default:
      NS_FATAL_ERROR("");
      return;;
    }

  m_packet->RemoveAtStart(bPro.getIdealMessageSize(recvData));
}

void
PeerConnection::HandlePeerClose(Ptr<Socket> socket)
{
  NS_LOG_FUNCTION(this);

  m_requestStrategy->StopRequest();

  m_socket->SetRecvCallback(
    MakeNullCallback<void, Ptr<Socket> >());
  m_connectionManager->RemoveFromPeerList(this);
}

void
PeerConnection::HandlePeerError(Ptr<Socket> socket)
{
  NS_LOG_FUNCTION(this);

}

//====================================================================================================

void
PeerConnection::SendChokeMessage()
{
  NS_LOG_FUNCTION(this);

  BitTorrentProtocol::GetInstance().SendChokeMessage(m_socket);
}

void
PeerConnection::SendUnchokeMessage()
{
  NS_LOG_FUNCTION(this);

  BitTorrentProtocol::GetInstance().SendUnchokeMessage(m_socket);
}

void
PeerConnection::SendHaveMessage(uint32_t pieceIndex)
{
  NS_LOG_FUNCTION(this);

  BitTorrentProtocol::GetInstance().SendHaveMessage(m_socket, pieceIndex);
}

void
PeerConnection::SendRequestMessage(uint32_t pieceIndex, uint32_t begin, uint32_t requestLength)
{
  NS_LOG_FUNCTION(this);

  if (BitTorrentProtocol::GetInstance().SendRequestMessage(m_socket, pieceIndex, begin, requestLength))
    {
      std::stringstream ss;
      ss << "Send a request";
      ss << " <pieceIndex: " << pieceIndex << " begin: " << begin << " requestLength: " << requestLength << ">";
      ss << " message.";
      PEER_CONNECTION_LOG_INFO(ss.str());
    }
}

void
PeerConnection::CloseConnection(void)
{
  m_requestStrategy->StopRequest();

  m_socket->Close();

  m_socket->SetRecvCallback(
    MakeNullCallback<void, Ptr<Socket> >());
}

std::string
PeerConnection::getPeerId(void)
{
  return m_peerInfo->getPeerId();
}

const Ptr<PeerInformation>&
PeerConnection::getPeerInformation(void)
{
  return m_peerInfo;
}

const Ptr<ConnectionManager>&
PeerConnection::getConnectionManager(void)
{
  return m_connectionManager;
}

//====================================================================================================

void
PeerConnection::RecvChokeMessage(std::string message)
{
  NS_LOG_FUNCTION(this);

  m_peerInfo->setUnchoking(false);
  m_requestStrategy->StopRequest();
}

void
PeerConnection::RecvUnchokeMessage(std::string message)
{
  NS_LOG_FUNCTION(this);

  m_peerInfo->setUnchoking(true);
  m_requestStrategy->StartRequest();
}

void
PeerConnection::RecvInterestedMessage(std::string message)
{
  NS_LOG_FUNCTION(this);

  m_peerInfo->setInteresting(true);
}

void
PeerConnection::RecvNotInterestedMessage(std::string message)
{
  NS_LOG_FUNCTION(this);

  m_peerInfo->setInteresting(false);
}

void
PeerConnection::RecvHaveMessage(std::string message)
{
  NS_LOG_FUNCTION(this);

  uint32_t pieceIndex;
  if (BitTorrentProtocol::GetInstance().ParseHaveMessage(message, &pieceIndex))
    {
      m_peerInfo->setBitfield(pieceIndex, true);
    }
}

void
PeerConnection::RecvBitfieldMessage(std::string message)
{
  NS_LOG_FUNCTION(this);

  std::string bitfield;
  if (BitTorrentProtocol::GetInstance().ParseBitfieldMessage(message, &bitfield))
    {
      m_peerInfo->setBitfield(bitfield);
    }
}

void
PeerConnection::RecvRequestMessage(Ptr<Socket> socket, std::string message)
{
  NS_LOG_FUNCTION(this);

  uint32_t pieceIndex;
  uint32_t begin;
  uint32_t requestLength;

  if (BitTorrentProtocol::GetInstance().ParseRequestMessage(message, &pieceIndex, &begin, &requestLength))
    {
      std::stringstream ss;
      ss << "Receive a request";
      ss << " <pieceIndex: " << pieceIndex << " begin: " << begin << " requestLength: " << requestLength << ">";
      ss << " message from [" << m_peerInfo->getPeerId() << "]";
      PEER_CONNECTION_LOG_INFO(ss.str());

      if (m_connectionManager->isUnchoking(this) &&
          m_connectionManager->getClientDatabase()->havePiece(pieceIndex))
        {
          std::string blockData;
          GenerateRandomString(&blockData, requestLength);
          if (BitTorrentProtocol::GetInstance().SendPieceMessage(socket, pieceIndex, begin, blockData))
            PEER_CONNECTION_LOG_INFO("Return a piece message");
        }
    }
}

void
PeerConnection::RecvPieceMessage(std::string message)
{
  NS_LOG_FUNCTION(this);

  uint32_t pieceIndex;
  uint32_t begin;
  std::string blockData;

  if (BitTorrentProtocol::GetInstance().ParsePieceMessage(message, &pieceIndex, &begin, &blockData))
    {
      std::stringstream ss;
      ss << "Receive piece";
      ss << " <pieceIndex: " << pieceIndex << " begin: " << begin << " DataLength: " << blockData.size() << ">";
      ss << " message from [" << m_peerInfo->getPeerId() << "]";
      PEER_CONNECTION_LOG_INFO(ss.str());

      m_connectionManager->ReceiveNewPiece(pieceIndex, begin, blockData);
    }
}

void
PeerConnection::RecvCancelMessage(std::string message)
{
  NS_LOG_FUNCTION(this);

  uint32_t pieceIndex;
  uint32_t begin;
  uint32_t requestLength;

  if (BitTorrentProtocol::GetInstance().ParseCancelMessage(message, &pieceIndex, &begin, &requestLength))
    {
      // UNIMPLEMENTED
    }
}

void
PeerConnection::RecvKeepAliveMessage(Ptr<Socket> socket, std::string message)
{
  NS_LOG_FUNCTION(this);

  if (m_waitKeepAliveResponse)
    {
      m_waitKeepAliveResponse = false;
    }
  else
    {
      BitTorrentProtocol::GetInstance().SendKeepAliveMessage(socket);
    }
}


//====================================================================================================

bool
PeerConnection::GenerateRandomString(std::string* str, uint32_t length)
{
  NS_LOG_FUNCTION(this);

  std::random_device rd;
  std::mt19937 mt(rd());
  std::uniform_int_distribution<uint8_t> dist(0, 0xFF);

  for (int i=0; i<length; i++)
    {
      str->append(1, dist(mt));
    }
}

void
PeerConnection::PEER_CONNECTION_LOG_INFO(std::string info)
{
  NS_LOG_INFO("[" << m_connectionManager->getClientDatabase()->getPeerId() << "] " <<
              Simulator::Now().GetSeconds() << "[s] " << info);
}



}

