/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */

#include "ns3/log.h"
#include "ns3/address-utils.h"
#include "ns3/socket.h"
#include "ns3/packet.h"
#include "ns3/simulator.h"

#include "ns3/BitTorrentClient.h"
#include "ns3/ClientDatabase.h"
#include "ns3/BitTorrentTools.h"
#include "ns3/BencodingObject.h"

#include "TrackerManager.h"

namespace ns3 {

NS_LOG_COMPONENT_DEFINE("TrackerManager");
NS_OBJECT_ENSURE_REGISTERED(TrackerManager);

TrackerManager::TrackerManager(Address trackerAddress, Ptr<BitTorrentClient> client, Ptr<ClientDatabase> clientDB)
{
  m_trackerAddress = trackerAddress;
  m_client         = client;
  m_clientDB       = clientDB;

  m_packet         = Create<Packet>();

  m_continue       = false;
}

bool
TrackerManager::StartRequest(void)
{
  m_nextAccess = Simulator::ScheduleNow(&TrackerManager::CreateConnection, this);
  m_continue   = true;
  return true;
}

bool
TrackerManager::StopRequest(void)
{
  m_continue = false;

  if (m_nextAccess.IsRunning())
    {
      Simulator::Cancel(m_nextAccess);
    }

  if (m_socket != 0)
    {
      m_socket->Close();
      m_socket->SetRecvCallback(MakeNullCallback<void, Ptr<Socket> >());
    }

  return true;
}

void
TrackerManager::CreateConnection(void)
{
  m_socket = m_client->CreateTcpSocket();
  
  if (InetSocketAddress::IsMatchingType(m_trackerAddress))
    {
      m_socket->Bind();
      m_socket->Connect(m_trackerAddress);
    }
  else if (Inet6SocketAddress::IsMatchingType(m_trackerAddress))
    {
      m_socket->Bind6();
      m_socket->Connect(m_trackerAddress);
    }
  
  m_socket->SetConnectCallback(
    MakeCallback(&TrackerManager::ConnectionSucceeded, this),
    MakeCallback(&TrackerManager::ConnectionFailed, this));
  m_socket->SetRecvCallback(
    MakeCallback(&TrackerManager::HandleReadResponse, this));
}

void
TrackerManager::ConnectionSucceeded(Ptr<Socket> socket)
{
  NS_LOG_FUNCTION(this);

  SendMessage();
}

void
TrackerManager::ConnectionFailed(Ptr<Socket> socket)
{
  NS_LOG_FUNCTION(this);

  //TODO: set next access time
}

int
TrackerManager::SendMessage(void)
{
  NS_LOG_FUNCTION(this);

  // Generate a Send message.
  std::stringstream ss;
  ss << "GET /announce?";
  ss << "info_hash="   << BitTorrentTools::UrlEncoder(m_clientDB->getInfoHash());
  ss << "&peer_id="    << BitTorrentTools::UrlEncoder(m_clientDB->getPeerId());
  ss << "&port="       << m_clientDB->listeningPort();
  ss << "&uploaded="   << m_clientDB->uploaded();
  ss << "&downloaded=" << m_clientDB->downloaded();
  ss << "&left="       << m_clientDB->left();

  switch (m_clientDB->getStatus())
    {
    case ClientDatabase::STARTED:
      ss << "&event=started";
      break;
    case ClientDatabase::STOPPED:
      ss << "&event=stopped";
      break;
    case ClientDatabase::COMPLETED:  
      ss << "&event=completed";
      break;
    }

  // Add ASN info
  ss << "&asn="        << m_clientDB->getAsn();

  ss << " HTTP/1.0";
  // End of the get method.

  ss << "\r\n\r\n";
  // End of the header.

  return m_socket->Send((uint8_t*)ss.str().data(), ss.str().size(), 0);
}

void
TrackerManager::HandleReadResponse(Ptr<Socket> socket)
{
  NS_LOG_FUNCTION(this);

  for (uint32_t available = socket->GetRxAvailable();
       available > 0; available = socket->GetRxAvailable())
    {
      m_packet->AddAtEnd(socket->Recv());
    }

  uint8_t *buf = new uint8_t[m_packet->GetSize()];
  m_packet->CopyData(buf, m_packet->GetSize());
  std::string recvData((char*)buf, m_packet->GetSize());
  delete[] buf;

  //TODO: packet size check.


  //main process
  Ptr<BencodingObject> recvObj;
  try
    {
      recvObj = BencodingObject::ParseBencodingObject((uint8_t*)recvData.data(), recvData.size());
    }
  catch (std::string e)
    {
      NS_LOG_INFO("parse error: " << e);
    }

  // Add PeerList to the ClientDB.
  try
    {
      m_clientDB->ClearPeerList();

      for (BencodingObject::listIterator it = recvObj->getValueFromDictionary("peers")->listBegin();
           it != recvObj->getValueFromDictionary("peers")->listEnd(); ++it)
        {
          std::string peer_id  = (*it)->getValueFromDictionary("peer_id")->getString();
          std::string ip       = (*it)->getValueFromDictionary("ip")->getString();
          std::string str_port = (*it)->getValueFromDictionary("port")->getString();
          uint16_t    port     = (uint16_t)strtol(str_port.c_str(), NULL, 10);
          std::string str_asn  = (*it)->getValueFromDictionary("asn")->getString();
          uint16_t    asn      = (uint16_t)strtol(str_asn.c_str(), NULL, 10);

          m_clientDB->AddToPeerList(peer_id, ip, port, asn);
        }
    }
  catch (std::string e)
    {
      NS_LOG_INFO("peers list error: " << e);
    }

  // Schedule next access
  if (m_continue)
    {
      try
        {
          int intervalSec = (int)(recvObj->getValueFromDictionary("interval")->getInteger());
          m_nextAccess    = Simulator::Schedule(Seconds(intervalSec), &TrackerManager::CreateConnection, this);
        }
      catch (std::string e)
        {
          NS_LOG_INFO("interval not found.");
        }
    }


  //TODO: erase 'used' data from packet.
  m_packet->RemoveAtStart(m_packet->GetSize());


}

}

