/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */

#include "ns3/log.h"
#include "ns3/simulator.h"

#include "ns3/BitTorrentProtocol.h"
#include "ns3/ClientDatabase.h"

#include "ConnectionStrategyBase.h"

namespace ns3 {

NS_LOG_COMPONENT_DEFINE("ConnectionStrategyBase");
NS_OBJECT_ENSURE_REGISTERED(ConnectionStrategyBase);

bool
ConnectionStrategyBase::isConnectable(const PeerInformation& pi)
{
  NS_LOG_FUNCTION(this);

  if (m_connectionManager->isConnected(pi.getPeerId()))
    {
      return false;
    }
  
  uint32_t restSocket = m_connectionManager->RestConnectableSocket() - m_connectedSocketList.size();
  if (!(restSocket > 0))
    {
      return false;
    }
  
  return true;
}

bool
ConnectionStrategyBase::Connect(PeerInformation& pi)
{
  NS_LOG_FUNCTION(this);

  if (!isConnectable(pi))
      return false;

  std::string ip   = pi.getIp();
  uint16_t    port = pi.getPort();

  InetSocketAddress address = InetSocketAddress(ip.c_str(), port);
  Ptr<Socket>       socket  = m_connectionManager->CreateTcpSocket();

  if (InetSocketAddress::IsMatchingType(address))
    {
      if (socket->Bind() == -1) false;
      if (socket->Connect(address) == -1) false;
    }

  std::stringstream ss;
  ss << "Connect to " << pi.getPeerId();
  CONNECTION_LOG_INFO(ss.str());  

  socket->SetConnectCallback(
    MakeCallback(&ConnectionStrategyBase::ConnectionSucceeded, this),
    MakeCallback(&ConnectionStrategyBase::ConnectionFailed, this));

  return true;
}

void
ConnectionStrategyBase::ConnectionSucceeded(Ptr<Socket> socket)
{
  NS_LOG_FUNCTION(this);

  std::stringstream ss;
  ss << " Connection Succeeded.";
  CONNECTION_LOG_INFO(ss.str());


  m_connectedSocketList[socket] = Create<Packet>();

  std::string peerId   = m_connectionManager->getClientDatabase()->getPeerId();
  std::string infoHash = m_connectionManager->getClientDatabase()->getInfoHash();

  BitTorrentProtocol::GetInstance().SendHandshakeMessage(socket, peerId, infoHash);

  socket->SetRecvCallback(
    MakeCallback(&ConnectionStrategyBase::HandleRead, this));
}

void
ConnectionStrategyBase::ConnectionFailed(Ptr<Socket> socket)
{
  NS_LOG_FUNCTION(this);
}

void
ConnectionStrategyBase::HandleRead(Ptr<Socket> socket)
{
  NS_LOG_FUNCTION(this);

  Ptr<Packet> packet = m_connectedSocketList[socket];

  for (uint32_t available = socket->GetRxAvailable();
       available > 0; available = socket->GetRxAvailable())
    {
      packet->AddAtEnd(socket->Recv());
    }

  uint8_t *buf = new uint8_t[packet->GetSize()];
  packet->CopyData(buf, packet->GetSize());
  std::string recvData((char*)buf, packet->GetSize());
  delete[] buf;

  //TODO: check the packet size
  // insert code here...

  BitTorrentProtocol bPro = BitTorrentProtocol::GetInstance();
  std::string peerId;
  std::string infoHash;

  if (bPro.ParseHandshakeMessage(recvData, &peerId, &infoHash) &&
      (infoHash.compare(m_connectionManager->getClientDatabase()->getInfoHash()) == 0))
    {
      EntrustPeerConnection(socket, peerId);
    }
  else
    {
      socket->Close();
      socket->SetRecvCallback(MakeNullCallback<void, Ptr<Socket> >());
      m_connectedSocketList.erase(socket);
    }
}

void
ConnectionStrategyBase::EntrustPeerConnection(Ptr<Socket> socket, std::string peerId)
{
  NS_LOG_FUNCTION(this);

  socket->SetRecvCallback(MakeNullCallback<void, Ptr<Socket> >());
  m_connectedSocketList.erase(socket);
  m_connectionManager->EntrustedPeerConnection(socket, peerId);

  std::stringstream ss;
  ss << "++++++++++ Handshake with [" << peerId << "] Succeeded!! (connection) ++++++++++";
  CONNECTION_LOG_INFO(ss.str());
}

void
ConnectionStrategyBase::CONNECTION_LOG_INFO(std::string info)
{
  NS_LOG_INFO("[" << m_connectionManager->getClientDatabase()->getPeerId() << "] " <<
              Simulator::Now().GetSeconds() << "[s] " << info);
}

}

