/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */

#include "BencodingObject.h"

#include <math.h>

namespace ns3{

Ptr<BencodingObject>
BencodingObject::ParseBencodingObject(uint8_t* data, const uint32_t length) throw (std::string)
{
  uint32_t cursor = 0;
  return ParseBencodingObject(data, length, &cursor);
}

Ptr<BencodingObject>
BencodingObject::ParseBencodingObject(uint8_t* data, const uint32_t length, uint32_t* cursor) throw (std::string)
{
  if (!(*cursor < length))
    {
      std::string error = "lack of data end.";
      throw error;
    }

  Ptr<BencodingObject> obj = NULL;

  if (data[*cursor] == 'i')
    {
      long integer = 0;
      (*cursor)++; // skip a letter of 'i'.

      for (; data[*cursor] != 'e'; (*cursor)++)
        {
          if (*cursor < length && '0' <= data[*cursor] && data[*cursor] <= '9')
            {
              integer = integer * 10 + (data[*cursor] - '0');
            }
          else if (!(*cursor < length))
            {
              std::string error = "lack of the tail data. (can't find integer end)";
              throw error;
            }
          else
            {
              std::string error = "unacceptable letter come. (integer type parse error)";
              throw error;
            }
        }

      (*cursor)++; // skip a letter of 'e' meaning the end.
      obj = Create<BencodingObject>(integer);

    }
  else if ( '1' <= data[*cursor] && data[*cursor] <= '9')
    {
      long stringLength = 0;

      for (; data[*cursor] != ':'; (*cursor)++) // parse string length
        {
          if (*cursor < length && '0' <= data[*cursor] && data[*cursor] <= '9')
            {
              stringLength = stringLength * 10 + (data[*cursor] - '0');
            }
          else if (!(*cursor < length))
            {
              std::string error = "can't find end of string type length.";
              throw error;
            }
          else
            {
              std::string error = "unacceptable letter come. (string type parse error)";
              throw error;
            }
        }

      (*cursor)++; // skip a letter of ':' meaning string length end.

      std::string string = "";

      for (; stringLength != 0; stringLength--) // parse string data
        {
          if ((*cursor) < length)
            {
              string += (char) data[(*cursor)++];
            }
          else
            {
              std::string error = "lack of string data. (string type parse error)";
              throw error;
            }
        }

      obj = Create<BencodingObject>(string);

    }
  else if (data[*cursor] == 'l')
    {
      obj = Create<BencodingObject>(BencodingObject::LIST);
      (*cursor)++; // skip a letter of 'l'.

      while (data[*cursor] != 'e')
        {
          if (*cursor < length)
            {
              obj->addToList(ParseBencodingObject(data, length, cursor));
            }
          else
            {
              std::string error = "can't find the end of list type.";
              throw error;
            }
        }

      (*cursor)++; // skip a letter of 'e' meaning list type end.

    }
  else if (data[*cursor] == 'd')
    {
      obj = Create<BencodingObject>(BencodingObject::DICTIONARY);
      (*cursor)++; // skip a letter of 'd'.

      while (data[*cursor] != 'e')
        {
          if (*cursor < length)
            {
              Ptr<BencodingObject> key = ParseBencodingObject(data, length, cursor);
              Ptr<BencodingObject> value = ParseBencodingObject(data, length, cursor);

              if (key->getObjectType() != BencodingObject::STRING)
                {
                  std::string error = "type of dictionary key is not BencodingObject::STRING";
                  throw error;
                }

              obj->addToDictionary(key->getString(), value);

            }
          else
            {
              std::string error = "can't find the end of dictionary type";
              throw error;
            }
        }

      (*cursor)++; // skip a letter of 'e' meaning dictionary type end.

    }
  else
    {
      std::string error = "the data type is unknown to Parser.";
      throw error;
    }

  return obj;
}

// ================================================================================
BencodingObject::BencodingObject() : m_type(BencodingObject::NULL_OBJECT)
{
}

BencodingObject::BencodingObject(BencodingObject::ObjectType type) : m_type(type)
{
  if (m_type == BencodingObject::LIST)
    {
      m_list = new std::list<Ptr<BencodingObject> > ();
    }
  else if (m_type == BencodingObject::DICTIONARY)
    {
      m_dictionary = new std::map<std::string, Ptr<BencodingObject> > ();
    }
}

BencodingObject::BencodingObject(long integer) : m_type(BencodingObject::INTEGER)
{
  m_integer = integer;
}

BencodingObject::BencodingObject(std::string string) : m_type(BencodingObject::STRING)
{
  m_string = string;
}

// ================================================================================

BencodingObject::~BencodingObject()
{
  clearObject();
}

void
BencodingObject::clearObject()
{
  switch (m_type)
    {
    case BencodingObject::INTEGER:
      m_integer = 0;
      break;
    case BencodingObject::STRING:
      m_string = "";
      break;
    case BencodingObject::LIST:
      delete m_list; //automatically release  Ptr<BencodingObject> ?
      m_list = NULL;
      break;
    case BencodingObject::DICTIONARY:
      delete m_dictionary; // automaticallly release Ptr<BencodingObject> ?
      m_dictionary = NULL;
      break;
    default:
      break;
    }

  m_type = BencodingObject::NULL_OBJECT;
}

// ================================================================================

void
BencodingObject::changeObjectType(BencodingObject::ObjectType newType)
{
  clearObject();
  m_type = newType;

  if (m_type == BencodingObject::LIST)
    {
      m_list = new std::list<Ptr<BencodingObject> > ();
    }
  else if (m_type == BencodingObject::DICTIONARY)
    {
      m_dictionary = new std::map<std::string, Ptr<BencodingObject> > ();
    }
}

BencodingObject::ObjectType
BencodingObject::getObjectType(void)
{
  return m_type;
}

// ================================================================================

void
BencodingObject::setInteger(long integer) throw (std::string)
{
  if (m_type == BencodingObject::INTEGER)
    {
      m_integer = integer;
    }
  else
    {
      std::string error = "This object isn't BencodingObject::INTEGER";
      throw error;
    }
}

long
BencodingObject::getInteger(void) throw (std::string)
{
  if (m_type == BencodingObject::INTEGER)
    {
      return m_integer;
    }
  else
    {
      std::string error = "This object isn't BencodingObject::INTEGER";
      throw error;
    }
}

void
BencodingObject::setString(std::string string) throw (std::string)
{
  if (m_type == BencodingObject::STRING)
    {
      m_string = string;
    }
  else
    {
      std::string error = "This object isn't BencodingObject::STRING";
      throw error;
    }
}

std::string
BencodingObject::getString(void) throw (std::string)
{
  if (m_type == BencodingObject::STRING)
    {
      return m_string;
    }
  else
    {
      std::string error = "This object isn't BencodingObject::STRING";
      throw error;
    }
}

void
BencodingObject::addToList(Ptr<BencodingObject> item) throw (std::string)
{
  if (m_type == BencodingObject::LIST)
    {
      m_list->push_back(item);
    }
  else
    {
      std::string error = "This object isn't BencodingObject::LIST";
      throw error;
    }
}

int
BencodingObject::getListSize(void) throw (std::string)
{
  if (m_type == BencodingObject::LIST)
    {
      return m_list->size();
    }
  else
    {
      std::string error = "This object isn't BencodingObject::LIST";
      throw error;
    }
}

BencodingObject::listIterator
BencodingObject::listBegin(void) throw (std::string)
{
  if (m_type == BencodingObject::LIST)
    {
      return m_list->begin();
    }
  else
    {
      std::string error = "This object isn't BencodingObject::LIST";
      throw error;
    }
}

BencodingObject::listIterator
BencodingObject::listEnd(void) throw (std::string)
{
  if (m_type == BencodingObject::LIST)
    {
      return m_list->end();
    }
  else
    {
      std::string error = "This object isn't BencodingObject::LIST";
      throw error;
    }
}

BencodingObject::listIterator
BencodingObject::eraseFromList(BencodingObject::listIterator it) throw (std::string)
{
  if (m_type == BencodingObject::LIST)
    {
      return m_list->erase(it);
    }
  else
    {
      std::string error = "This object isn't BencodingObject::LIST";
      throw error;
    }
}

void
BencodingObject::addToDictionary(std::string key, Ptr<BencodingObject> value) throw (std::string)
{
  if (m_type == BencodingObject::DICTIONARY)
    {
      (*m_dictionary)[key] = value;
    }
  else
    {
      std::string error = "This object isn't BencodingObject::DICTIONARY";
      throw error;
    }
}

int
BencodingObject::getDictionarySize(void) throw (std::string)
{
  if (m_type == BencodingObject::DICTIONARY)
    {
      return m_dictionary->size();
    }
  else
    {
      std::string error = "This object isn't BencodingObject::DICTIONARY";
      throw error;
    }
}


BencodingObject::dictionaryIterator
BencodingObject::dictionaryBegin(void) throw (std::string)
{
  if (m_type == BencodingObject::DICTIONARY)
    {
      return m_dictionary->begin();
    }
  else
    {
      std::string error = "This object isn't BencodingObject::DICTIONARY";
      throw error;
    }
}

BencodingObject::dictionaryIterator
BencodingObject::dictionaryEnd(void) throw (std::string)
{
  if (m_type == BencodingObject::DICTIONARY)
    {
      return m_dictionary->end();
    }
  else
    {
      std::string error = "This object isn't BencodingObject::DICTIONARY";
      throw error;
    }
}

BencodingObject::dictionaryIterator
BencodingObject::eraseFromDictionary(BencodingObject::dictionaryIterator it) throw (std::string)
{
  if (m_type == BencodingObject::DICTIONARY)
    {
      return m_dictionary->erase(it);
    }
  else
    {
      std::string error = "This object isn't BencodingObject::DICTIONARY";
      throw error;
    }
}

Ptr<BencodingObject>
BencodingObject::getValueFromDictionary(std::string key) throw (std::string)
{
  if (m_type == BencodingObject::DICTIONARY)
    {
      return (*m_dictionary)[key];
    }
  else
    {
      std::string error = "This object isn't BencodingObject::DICTIONARY";
      throw error;      
    }
}



// ================================================================================

uint32_t
BencodingObject::serialize(uint8_t* buf, const uint32_t length) throw (std::string)
{
  uint32_t bytes = 0;

  switch (m_type)
    {
    case BencodingObject::INTEGER:
      {
        char temp[21] = {'\0'}; // ((int)log10(2^63) + 1) + 1(signed/unsigned) + 1('\0') = 21 bytes
        snprintf(temp, sizeof(temp)/sizeof(char), "%ld", m_integer);

        uint32_t writeSize = 2; // start letter 'i' and end letter 'e'.
        for (int i=0; temp[i] != '\0'; i++) // string length of integer.
          {
            writeSize++;
          }

        if (writeSize > length)
          {
            std::string error = "buffer size is too small.";
            throw error;
          }

        buf[bytes++] = 'i';
        for (int i=0; temp[i] != '\0'; i++)
          {
            buf[bytes++] = (uint8_t) temp[i];
          }
        buf[bytes++] = 'e';

        break;
      }
    case BencodingObject::STRING:
      {
        uint32_t    stringLength = (uint32_t) m_string.length();
        const char* string       = m_string.c_str();

        uint32_t writeSize = ((int)log10(stringLength) + 1) + 1 + stringLength; // string length + ':' + string data
        if (writeSize > length)
          {
            std::string error = "buffer size is too small.";
            throw error;
          }

        char temp[10] = {'\0'}; // ((int)log10(2^23) + 1) + 1('\0') = 10 bytes
        snprintf(temp, sizeof(temp)/sizeof(char), "%u", stringLength);

        for (int i=0; temp[i] != '\0'; i++)
          {
            buf[bytes++] = temp[i];
          }

        buf[bytes++] = ':';

        for (int i=0; i<stringLength; i++)
          {
            buf[bytes++] = string[i];
          }

        break;
      }
    case BencodingObject::LIST:
      {
        if (bytes + 1 > length) // list start 'l'.
          {
            std::string error = "buffer size is too small.";
            throw error;
          }
        else
          {
            buf[bytes++] = 'l';
          }

        for (std::list<Ptr<BencodingObject> >::iterator it = m_list->begin(); it!=m_list->end(); it++)
          {
            bytes += (*it)->serialize(buf + bytes, length - bytes);
          }

        if (bytes + 1 > length) // list end 'e'.
          {
            std::string error = "buffer size is too small.";
            throw error;
          }
        else
          {
            buf[bytes++] = 'e';
          }

        break;
      }
    case BencodingObject::DICTIONARY:
      {
        if (bytes + 1 > length) // dictionary start 'd'.
          {
            std::string error = "buffer size is too small.";
            throw error;
          }
        else
          {
            buf[bytes++] = 'd';
          }

        for (std::map<std::string, Ptr<BencodingObject> >::iterator it = m_dictionary->begin(); it!=m_dictionary->end(); it++)
          {
            std::string key = it->first;
            Ptr<BencodingObject> value = it->second;

            // dictionary key write to buffer
            uint32_t    keyStringLength = (uint32_t) key.length();
            const char* keyString       = key.c_str();

            uint32_t writeSize = ((uint32_t)log10(keyStringLength) + 1) + 1 + keyStringLength; // keyStringlength + ':' + keyString [bytes]

            if (writeSize > length)
              {
                std::string error = "buffer size is too small.";
                throw error;
              }

            char temp[10] = {'\0'}; // ((int)log10(2^32) + 1) + 1('\0')
            snprintf(temp, sizeof(temp)/sizeof(char), "%u", keyStringLength);

            for (int i=0; temp[i] != '\0'; i++)
              {
                buf[bytes++] = temp[i];
              }

            buf[bytes++] = ':';

            for (int i=0; i<keyStringLength; i++)
              {
                buf[bytes++] = keyString[i];
              }

            // dictionary value write to buffer
            bytes += value->serialize(buf + bytes, length - bytes);
          }

        if (bytes + 1 > length)
          {
            std::string error = "buffer size is too small.";
            throw error;
          }
        else
          {
            buf[bytes++] = 'e';
          }

        break;
      }
    default:
      {
        std::string error = "can't serialize this object";
        throw error;
        break;
      }
    }

  return bytes;
}

}
