/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */

#include <algorithm>

#include "ns3/log.h"
#include "ns3/ipv4-address.h"
#include "ns3/address-utils.h"
#include "ns3/inet-socket-address.h"
#include "ns3/socket.h"
#include "ns3/simulator.h"
#include "ns3/socket-factory.h"
#include "ns3/tcp-socket-factory.h"
#include "ns3/packet.h"
#include "ns3/ipv4.h"

#include "ns3/BitTorrentTools.h"
#include "ns3/BitTorrentTracker.h"

namespace ns3 {

NS_LOG_COMPONENT_DEFINE("BitTorrentTracker");
NS_OBJECT_ENSURE_REGISTERED(BitTorrentTracker);

// ==================================================
//  Public Member Functions
// ==================================================

BitTorrentTracker::BitTorrentTracker() : m_listeningPort(6969),
                           m_socket(0)
{
  NS_LOG_FUNCTION(this);

  m_peerList = Create<BencodingObject>(BencodingObject::DICTIONARY);
  m_peerList->addToDictionary("interval", Create<BencodingObject>(5));
  m_peerList->addToDictionary("peers", Create<BencodingObject>(BencodingObject::LIST));
}

BitTorrentTracker::~BitTorrentTracker()
{
  NS_LOG_FUNCTION(this);
}

void
BitTorrentTracker::SetListeningPort(uint16_t port)
{
  NS_LOG_FUNCTION(this);
  m_listeningPort = port;
}

// ==================================================
//  Protected Member Functions
// ==================================================

void
BitTorrentTracker::DoDispose(void)
{
  NS_LOG_FUNCTION(this);
  m_socket = 0;
  m_socketList.clear();

  Application::DoDispose();
}

// ==================================================
//  Private Member Functions
// ==================================================

void
BitTorrentTracker::StartApplication(void)
{
  NS_LOG_FUNCTION(this);

  if (!m_socket)
    {
      m_socket = Socket::CreateSocket(GetNode(), TcpSocketFactory::GetTypeId());

      Ptr<Ipv4> ipv4List = GetNode()->GetObject<Ipv4>();
      Ipv4Address listeningAddress = ipv4List->GetAddress(/* Interface index */ 1, /* Address index */ 0).GetLocal();
      InetSocketAddress listeningAddressWithPort = InetSocketAddress(listeningAddress, m_listeningPort);

      uint32_t ipv4 = listeningAddress.Get();
      NS_LOG_INFO("Listening Address: " << ((ipv4 >> 24) & 0xff) << "."
                                        << ((ipv4 >> 16) & 0xff) << "."
                                        << ((ipv4 >>  8) & 0xff) << "."
                                        << (ipv4 &0xff)          << " "
                            << "Port: " << m_listeningPort       );

      m_socket->Bind(listeningAddressWithPort);
      m_socket->Listen();
    }

  m_socket->SetRecvCallback(
    MakeCallback(&BitTorrentTracker::HandleRead, this));
  m_socket->SetAcceptCallback(
    MakeNullCallback<bool, Ptr<Socket>, const Address &>(),
    MakeCallback(&BitTorrentTracker::HandleAccept, this));
  m_socket->SetCloseCallbacks(
    MakeCallback(&BitTorrentTracker::HandlePeerClose, this),
    MakeCallback(&BitTorrentTracker::HandlePeerError, this));
}

void
BitTorrentTracker::StopApplication(void)
{
  NS_LOG_FUNCTION(this);
  while(!m_socketList.empty())
    {
      Ptr<Socket> acceptedSocket = m_socketList.front();
      m_socketList.pop_front();
      acceptedSocket->Close();
    }

  if (m_socket)
    {
      m_socket->Close();
      m_socket->SetRecvCallback(MakeNullCallback<void, Ptr<Socket> >());
    }
}

void
BitTorrentTracker::HandleRead(Ptr<Socket> socket)
{
  NS_LOG_FUNCTION(this);
  
  char buf[1024] = {'\0'}; // 1024byte
  Address recvFrom;
  int recvSize = socket->RecvFrom((uint8_t*)buf, sizeof(buf)/sizeof(char), 0, recvFrom);

  std::string recvMessage = buf;

  if (recvMessage.find("GET ", 0) != 0) // Not Find GET request header.
    {
      NS_FATAL_ERROR("The tracker server received a garbage data.");

      std::stringstream ss;
      ss << "HTTP/1.0 400 Bad Request\r\n\r\n";
      ss << "You sent me a garbage data.";

      socket->Send((uint8_t*)ss.str().data(), ss.str().size(), 0);
      socket->Close();

      return;
    }

  // Parse GET request Query
  std::map<std::string, std::string> queryDict;

  std::string::iterator queryStart = std::find(recvMessage.begin(), recvMessage.end(), '?');
  std::string::iterator queryEnd = std::find(queryStart, recvMessage.end(), ' ');

  std::string queryMessage(queryStart + 1, queryEnd); // skip a letter of '?'
  std::string::size_type pos = 0;

  while(1)
    {
      std::string::size_type equalPos = queryMessage.find('=', pos);
      std::string::size_type splitPos = queryMessage.find('&', pos);

      std::string queryKey(queryMessage, pos, equalPos - pos);
      std::string queryValue(queryMessage, equalPos + 1, splitPos - equalPos - 1); // '+1/-1' remove a letter of '=/&'.
      
      if (queryKey.compare("info_hash") == 0 || queryKey.compare("peer_id") == 0)
        {
          queryValue = BitTorrentTools::UrlDecoder(queryValue);
        }

      queryDict[queryKey] = queryValue;
      //      NS_LOG_INFO("queryMessage Key=\"" << queryKey << "\" Value=\"" << queryValue << "\"");

      if (splitPos != std::string::npos)
        pos = ++splitPos;
      else
        break;
    }

  // Convert sender ipv4 Address
  std::string senderIpAddress;
  if (InetSocketAddress::IsMatchingType(recvFrom))
    {
      InetSocketAddress inet = InetSocketAddress::ConvertFrom(recvFrom);
      Ipv4Address from = inet.GetIpv4();
      uint32_t ipv4Addr = from.Get();

      std::stringstream ss;
      ss << ((ipv4Addr >> 24) & 0xff) << "."
         << ((ipv4Addr >> 16) & 0xff) << "."
         << ((ipv4Addr >>  8) & 0xff) << "."
         << (ipv4Addr & 0xff);
      senderIpAddress = ss.str();
    }
  else
    {
      NS_FATAL_ERROR("Sender IP is not ipv4.");
    }

  NS_LOG_INFO("Sender Ipv4: " << senderIpAddress);
  NS_LOG_INFO(Simulator::Now().GetSeconds() << "[s] received message. " << recvSize << "[bytes]\n" << recvMessage);


  if (queryDict["event"].compare("started")   == 0 ||
      queryDict["event"].compare("completed") == 0)
    {
      AddToPeerList(queryDict["peer_id"], senderIpAddress, queryDict["port"], queryDict["asn"]);
      
      // Return Bencoded peer info data.
      SendPeerList(socket);
    }
  else if (queryDict["event"].compare("stopped") == 0)
    {
      RemoveFromPeerList(queryDict["peer_id"]);
    }
}

void
BitTorrentTracker::HandleAccept(Ptr<Socket> socket, const Address& from)
{
  NS_LOG_FUNCTION(this << socket << from);

  socket->SetRecvCallback(MakeCallback(&BitTorrentTracker::HandleRead, this));
  m_socketList.push_back(socket);
}

void
BitTorrentTracker::HandlePeerClose(Ptr<Socket> socket)
{
  NS_LOG_FUNCTION(this);
}

void
BitTorrentTracker::HandlePeerError(Ptr<Socket> socket)
{
  NS_LOG_FUNCTION(this);
}

void
BitTorrentTracker::AddToPeerList(std::string peerId, std::string ipAddress, std::string port, std::string asn)
{
  NS_LOG_FUNCTION(this);

  // remove same "peer_id" info from the List.
  RemoveFromPeerList(peerId);

  // add new "peer_id" info to the List.
  Ptr<BencodingObject> peersList = m_peerList->getValueFromDictionary("peers");
  Ptr<BencodingObject> peerInfo = Create<BencodingObject>(BencodingObject::DICTIONARY);
  peerInfo->addToDictionary("peer_id", Create<BencodingObject>(peerId));
  peerInfo->addToDictionary("ip", Create<BencodingObject>(ipAddress));
  peerInfo->addToDictionary("port", Create<BencodingObject>(port));
  peerInfo->addToDictionary("asn", Create<BencodingObject>(asn));
  peersList->addToList(peerInfo);

  NS_LOG_INFO("Add to the DB: " << peerId);
}

void
BitTorrentTracker::RemoveFromPeerList(std::string peerId)
{
  NS_LOG_FUNCTION(this);

  Ptr<BencodingObject> peersList = m_peerList->getValueFromDictionary("peers");

  for (BencodingObject::listIterator it = peersList->listBegin(); it != peersList->listEnd(); ++it)
    {
      std::string existedPeerId = (*it)->getValueFromDictionary("peer_id")->getString();
      if (peerId.compare(existedPeerId) == 0)
        {
          NS_LOG_INFO("Remove from the DB: " << existedPeerId);
          peersList->eraseFromList(it);
          break;
        }
    }
}

void
BitTorrentTracker::SendPeerList(Ptr<Socket> socket)
{
  NS_LOG_FUNCTION(this);

  uint8_t buf[1024 * 1024] = {'\0'}; // 1024 kbyte
  uint32_t dataSize = m_peerList->serialize(buf, sizeof(buf)/sizeof(uint8_t));

  int status = socket->Send(buf, dataSize, 0);

  NS_LOG_INFO(Simulator::Now().GetSeconds() << "[s] send message. " << dataSize << "[bytes]\n" << buf);
}

void
BitTorrentTracker::SendFailureMessage(Ptr<Socket> socket)
{
  NS_LOG_FUNCTION(this);

  Ptr<BencodingObject> failureMessageObject = Create<BencodingObject>(BencodingObject::DICTIONARY);
  failureMessageObject->addToDictionary("failure reason", Create<BencodingObject>("you send me garbage data."));

  uint8_t buf[512] = {'\0'}; // 512 byte
  uint32_t dataSize = failureMessageObject->serialize(buf, sizeof(buf)/sizeof(uint8_t));

  int status = socket->Send(buf, dataSize, 0);

  NS_LOG_INFO("SendMessage: " << buf);
}



} // namespace ns3

